﻿
// The sole owner of this code is Dundas Data Visualization Inc.
// Copyright (c) 2015 Dundas Data Visualization Inc.
// All rights reserved.

using Dundas.BI;
using Dundas.BI.AccountServices;
using Dundas.BI.AccountServices.Extensibility;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace MyCompany.Sample.CustomAccountsProviderSample
{
	/// <summary>
	/// This class represents a simple accounts provider that uses a csv file. 
	/// </summary>
	public class CustomAccountProvider : IAccountsProvider2
	{
        #region Fields

        private string _accountInformationFilePath;
        private string _passwordInformationFilePath;

        #endregion Fields

        #region Internal Properties

        /// <summary>
        /// Gets the account information file path.
        /// </summary>
        /// <value>
        /// The account information file path.
        /// </value>
        internal string PasswordInformationFilePath
        {
            get
            {
                if (_passwordInformationFilePath == null)
                {
                    if (System.Web.HttpContext.Current == null)
                    {
                        return null;
                    }

                    _passwordInformationFilePath =
                        System.Web.HttpContext.Current.Server.MapPath("/App_Data/PasswordInformation.csv");
                }

                return _passwordInformationFilePath;
            }
        }

        /// <summary>
        /// Gets the account information file path.
        /// </summary>
        /// <value>
        /// The account information file path.
        /// </value>
        internal string AccountInformationFilePath
        {
	        get
	        {
		        if(_accountInformationFilePath == null)
		        {
                    if(System.Web.HttpContext.Current == null)
                    {
                        return null;
                    }

			        _accountInformationFilePath =
				        System.Web.HttpContext.Current.Server.MapPath("/App_Data/AccountInformation.csv");
		        }

		        return _accountInformationFilePath;
	        }
        }

        #endregion Internal Properties

        #region Public Properties
        
        /// <summary>
        /// Gets a value indicating whether modify is supported.
        /// </summary>
        /// <value>
        /// <c>true</c> if modify is supported; otherwise, <c>false</c>.
        /// </value>
        public bool IsModifySupported
        {
	        get { return true; }
        }

        /// <summary>
        /// Gets the supported query operations.
        /// </summary>
        /// <value>
        /// The supported query operations.
        /// </value>
        public Dundas.BI.Utility.QueryApiOperations SupportedQueryOperations
        {
	        get { return Dundas.BI.Utility.QueryApiOperations.Count; }
        }

        #endregion Public Properties

        #region Private Methods

        private string GetPassword(Guid accountId)
        {
            return GetAllPasswords()[accountId];
        }

        private void SaveAllPasswordsToFile(IDictionary<Guid, string> passwords)
        {
            StringBuilder passwordsFileContent = new StringBuilder();

            foreach(KeyValuePair<Guid,string> pair in passwords)
            {
                passwordsFileContent.AppendFormat(
                    CultureInfo.CurrentCulture,
                    "{0},{1}{2}",
                    pair.Key.ToString(),
                    pair.Value.ToString(),
                    Environment.NewLine
                    );
            }

            File.WriteAllText(this.PasswordInformationFilePath, passwordsFileContent.ToString());
        }

        /// <summary>
        /// Saves the account data to file.
        /// </summary>
        /// <param name="accountDataList">The account data list.</param>
        private void SaveAllAccountDataToFile(IEnumerable<AccountData> accountDataList)
        {
            StringBuilder updatedFileContent = new StringBuilder();

            foreach (LocalUserAccountData accountData in accountDataList)
            {
                updatedFileContent.AppendFormat(
                    CultureInfo.CurrentCulture,
                    "{0},{1},{2},{3},{4},{5},{6},{7}{8}",
                    accountData.Id,
                    accountData.Name,
                    accountData.IsEnabled,
                    accountData.FailedLogOnCount,
                    accountData.LockedUntil,
                    accountData.PasswordNeverExpires,
                    accountData.PasswordExpiryDate,
                    accountData.AccountExpiryDate,
                    Environment.NewLine
                    );
            }

            File.WriteAllText(this.AccountInformationFilePath, updatedFileContent.ToString());
        }

        private void UpdatePasswordIfNecessary(AccountData record)
        {
            LocalUserAccountData localUserAccountData = (LocalUserAccountData)record;

            if (localUserAccountData.Password != null)
            {
                SavePassword(localUserAccountData.Id, localUserAccountData.Password);
            }
        }

        #endregion Private Methods

        #region Public Methods


        /// <summary>
        /// Deletes the record.
        /// </summary>
        /// <param name="recordId">The record identifier.</param>
        /// <exception cref="System.NotImplementedException"></exception>
        public void DeleteRecord(Guid recordId)
        {
            SaveAllAccountDataToFile(
                GetAllRecords().Where(
                    record => !record.Id.Equals(recordId)
                    )
            );
        }

        /// <summary>
        /// Export an account from the underlying storage mechanism.
        /// </summary>
        /// <param name="accountId">The ID of the account to be exported.</param>
        /// <returns>
        /// The account data to be exported.
        /// </returns>
        public AccountData ExportRecord(Guid accountId)
        {
	        return GetAllRecords().FirstOrDefault(
		        accountDataItem => accountDataItem.Id.Equals(
			        accountId
			        )
		        );
        }

        public IDictionary<Guid,string> GetAllPasswords()
        {
            Dictionary<Guid, string> passwords = new Dictionary<Guid, string>();

            // Passwords
            foreach (string line in File.ReadAllLines(
                this.PasswordInformationFilePath
                ))
            {
                string[] splitLine =
                    line.Split(',');

                passwords.Add(new Guid(splitLine[0]), splitLine[1]);
                
            }

            return passwords;
        }

        public void SavePassword(Guid accountId, string password)
        {
            IDictionary<Guid, string> passwords = GetAllPasswords();

            passwords[accountId] = password;

            SaveAllPasswordsToFile(passwords);

        }



        /// <summary>
        /// Gets all user records.
        /// </summary>
        /// <returns>A IEnumerable of <see cref="T:Dundas.BI.AccountServices.Extensibility.AccountData" />.</returns>
        public IEnumerable<AccountData> GetAllRecords()
        {
            if(string.IsNullOrEmpty(this.AccountInformationFilePath))
            {
                return null;
            }

	        List<AccountData> accountDataList =
		        new List<AccountData>();

	        foreach(string line in File.ReadAllLines(
		        this.AccountInformationFilePath
		        ))
	        {
		        string[] splitLine =
			        line.Split(',');

                DateTime? lockedUntil =
                    null;

                if(!string.IsNullOrEmpty(splitLine[4]))
                {
                    lockedUntil =
                        DateTime.Parse(splitLine[4]);
                }

                DateTime? passwordExpiryDate = null;

                if (!string.IsNullOrEmpty(splitLine[6]))
                {
                    passwordExpiryDate =
                        DateTime.Parse(splitLine[6]);
                }

                DateTime? accountExpiryDate = null;

                if (!string.IsNullOrEmpty(splitLine[7]))
                {
                    accountExpiryDate =
                        DateTime.Parse(splitLine[7]);
                }

                accountDataList.Add(
			        new LocalUserAccountData()
				        {
                                Id = new Guid(splitLine[0]),
                                Name = splitLine[1],
                                IsEnabled = bool.Parse(splitLine[2]),
                                CanChangePassword = true,
                                FailedLogOnCount = int.Parse(splitLine[3]),
                                LockedUntil = lockedUntil,
                                PasswordNeverExpires = bool.Parse(splitLine[5]),
                                PasswordExpiryDate = passwordExpiryDate,
                                AccountExpiryDate = accountExpiryDate

                    }
			        );
	        }

	        return accountDataList;
        }


        /// <summary>
        /// Gets a collection of <see cref="T:Dundas.BI.AccountServices.Extensibility.AccountData" />.
        /// </summary>
        /// <param name="recordIds">The record ids.</param>
        /// <returns>A collection of <see cref="T:Dundas.BI.AccountServices.Extensibility.AccountData" />.</returns>
        public ICollection<AccountData> GetById(ICollection<Guid> recordIds)
        {
	        Collection<AccountData> accountDataCollection =
		        new Collection<AccountData>();

	        var allRecords =
		        GetAllRecords();

	        foreach(Guid id in recordIds)
	        {
                AccountData foundItem = 
                    allRecords.FirstOrDefault(
                        accountDataItem => accountDataItem.Id.Equals(id)
                    );

                if (foundItem != null)
                {
                    accountDataCollection.Add(
                        foundItem
                        );
                }
	        }

	        return accountDataCollection;
        }

        /// <summary>
        /// Retrieves an account specified by its name.
        /// </summary>
        /// <param name="accountName">The account's name.</param>
        /// <returns>
        /// The <see cref="T:Dundas.BI.AccountServices.Extensibility.AccountData" /> or <see langword="null" /> if it does not exist.
        /// </returns>
        public AccountData GetByName(string accountName)
        {
	        return GetAllRecords().FirstOrDefault(
		        accountDataItem => accountDataItem.Name.Equals(
			        accountName, StringComparison.CurrentCultureIgnoreCase
			        )
		        );
        }



        /// <summary>
        /// Imports a record to the underlying storage mechanism.
        /// </summary>
        /// <param name="record">The record to be imported.</param>
        /// <returns><c>true</c> if a new record was imported; otherwise, <c>false</c>.</returns>
        public bool ImportRecord(AccountData record)
        {

            IEnumerable<AccountData> allAccounts = 
                GetAllRecords();

            AccountData accountData =
                allAccounts.FirstOrDefault(
                    accountItem =>
                        accountItem.Id.Equals(record.Id)
                    );

            UpdatePasswordIfNecessary(record);

            if (accountData != null)
            {
                // Overwrite existin record.
                accountData = record;

                SaveAllAccountDataToFile(allAccounts);

                return false;
            }
            else
            {
                // New record.
                SaveAllAccountDataToFile(
                    allAccounts.Concat(new AccountData[] { record })
                    );
                return true;
            }


        }



        /// <summary>
        /// This operation is not supported by this account provider.
        /// </summary>
        /// <exception cref="System.NotSupportedException"></exception>
        public IList<Guid> Query(
            int pageNumber, int pageSize, IList<Tuple<Dundas.BI.AccountServices.AccountQueryField, 
            Dundas.BI.SortDirection>> orderBy, ICollection<Dundas.BI.AccountServices.AccountQueryFilterRule> filter)
        {
	        throw new NotSupportedException();
        }

        /// <summary>
        /// Queries the amount of users.
        /// </summary>
        /// <param name="filter">The filter.</param>
        /// <returns>The number of accounts.</returns>
        public int QueryCount(ICollection<Dundas.BI.AccountServices.AccountQueryFilterRule> filter)
        {
	        return GetAllRecords().Count();
        }

        /// <summary>
        /// Saves the record.
        /// </summary>
        /// <param name="record">The record.</param>
        /// <returns>A Guid representing the id of the item it just saved.</returns>
        public Guid SaveRecord(AccountData record)
        {
    
            AccountData[] listToSave;

            if(record.Id.Equals(Guid.Empty))
            {
                // Record is new
                record.Id = Guid.NewGuid();
                listToSave =
                    GetAllRecords().Concat(
                        new AccountData[] { record }
                    ).ToArray();
            }
            else
            {
                // Existing record
                listToSave = 
                    GetAllRecords().ToArray();

                for (int index = 0; index < listToSave.Count(); index++)
                {
                    if (listToSave[index].Id.Equals(record.Id))
                    {
                        listToSave[index] =
                            record;

                        break;
                    }
                }   
            }

            LocalUserAccountData localUserAccountData = (LocalUserAccountData)record;

            if(!string.IsNullOrEmpty(localUserAccountData.Password))
            {
                SavePassword(localUserAccountData.Id, localUserAccountData.Password);
            }

            UpdatePasswordIfNecessary(record);

            SaveAllAccountDataToFile(listToSave);

	        return record.Id;

	        
        }

        /// <summary>
        /// Updates a local user account's password.
        /// </summary>
        /// <param name="accountId">The ID of the account.</param>
        /// <param name="newPassword">The new password.</param>
        public void SetLocalUserAccountNewPassword(Guid accountId, string newPassword)
        {
            SavePassword(accountId, newPassword);
            
        }

        /// <summary>
        /// Updates the logon timestamp and logon count of a <see cref="T:Dundas.BI.AccountServices.LocalUserAccount" />
        /// or a <see cref="T:Dundas.BI.AccountServices.WindowsUserAccount" />.
        /// </summary>
        /// <param name="accountId">The ID of the account to update.</param>
        /// <param name="logOnTimestamp">The logon timestamp (UTC).</param>
        public void UpdateLogOnTimestampAndCount(Guid accountId, DateTime logOnTimestamp)
        {
            // This method is not called in IAccountsProvider2, is replaced by UpdateDynamicAccountProperties
        }


        /// <summary>
        /// Populates the local log on context.
        /// </summary>
        /// <param name="context">The context.</param>
        public void PopulateLocalLogOnContext(LocalLogOnContext context)
        {
           

            // Param validation.
            if(context == null)
            {
                throw new ArgumentNullException("context");
            }

            if (context.Credentials.Any(c => c.Id == LogOnCredentialIds.AccountName) == false
                ||
                context.Credentials.Any(c => c.Id == LogOnCredentialIds.Password) == false)
            {
                // No username or password provided.
                return;
            }

            LogOnCredential usernameCredential = context.Credentials.First(c => c.Id == LogOnCredentialIds.AccountName);
            LogOnCredential passwordCredential = context.Credentials.First(c => c.Id == LogOnCredentialIds.Password);

            IEnumerable<AccountData> accountDataList =
                GetAllRecords();

            AccountData accountData =
                accountDataList.FirstOrDefault(
                    accountDataItem => 
                        accountDataItem.Name.Equals(usernameCredential.Value)
                    );

            if(accountData == null)
            {
                // Account does not exist.
                return;
            }

            LocalUserAccountData localUserAccount = 
                (LocalUserAccountData)accountData;

            // Populate log on context.
            context.AccountId = localUserAccount.Id;
            context.LockedUntil = localUserAccount.LockedUntil;
            context.FailedLogOnCount = localUserAccount.FailedLogOnCount;
            
        }

        /// <summary>
        /// Resets any information about failed logon attempts for the specified account.
        /// </summary>
        /// <param name="accountId">The ID of the account.</param>
        public void ResetFailedLogOnInfo(Guid accountId)
        {
            IEnumerable<AccountData> accountDataList =
                GetAllRecords();

            LocalUserAccountData accountData =
                (LocalUserAccountData)accountDataList.First(accountDataItem => accountDataItem.Id.Equals(accountId));

            accountData.FailedLogOnCount =
                0;

            SaveAllAccountDataToFile(accountDataList);
        }

        /// <summary>
        /// Updates the log on failure information.
        /// </summary>
        /// <param name="accountId">The ID of the account.</param>
        /// <param name="failedLogOnCount">The new number of consecutive failed logon attempts.</param>
        /// <param name="lockedUntil">The time until which the account is locked, or <see langword="null" /> if the account is not
        /// yet locked.</param>
        public void UpdateLogOnFailureInfo(Guid accountId, int failedLogOnCount, DateTime? lockedUntil)
        {

            // This method is not called in IAccountsProvider2, is replaced by UpdateDynamicAccountProperties
        }

        /// <summary>
        /// Validates local user account logon credentials.
        /// </summary>
        /// <param name="context">The logon context.</param>
        /// <returns>
        ///   <see langword="true" /> if the credentials are valid; otherwise, <see langword="false" />.
        /// </returns>
        /// <exception cref="System.ArgumentNullException">context</exception>
        public bool ValidateLocalUserLogOnCredentials(LocalLogOnContext context)
        {
            

            // Param validation.
            if (context == null)
            {
                throw new ArgumentNullException("context");
            }

            LogOnCredential usernameCredential = context.Credentials.First(c => c.Id == LogOnCredentialIds.AccountName);
            LogOnCredential passwordCredential = context.Credentials.First(c => c.Id == LogOnCredentialIds.Password);

            IEnumerable<AccountData> accountDataList =
                GetAllRecords();

            AccountData accountData =
                accountDataList.FirstOrDefault(
                    accountDataItem =>
                        accountDataItem.Name.Equals(usernameCredential.Value)
                );

            LocalUserAccountData userAccount = 
                (LocalUserAccountData)accountData;

           

            if (passwordCredential.Value.Equals(GetPassword(accountData.Id)))
            {
                // Nothing went wrong - return success.
                return true;
            }
            else
            {               
                // Passwords did not match.
                return false;

            }

        }

        /// <summary>
        /// Update the dynamic properties of an account.
        /// </summary>
        /// <param name="accountId">The ID of the account to update.</param>
        /// <param name="properties">Dynamic properties of the account.</param>
        public void UpdateDynamicAccountProperties(Guid accountId, DynamicAccountProperties properties)
        {

            AccountData[] accountDataList =
                GetAllRecords().ToArray();

            for (int index = 0; index < accountDataList.Count(); index++)
            {
                if (accountDataList[index].Id.Equals(accountId))
                {
                    LocalUserAccountData accountData =
                        (LocalUserAccountData)accountDataList[index];

                    if (properties.FailedLogOnCount != null)
                    {
                        accountData.FailedLogOnCount = properties.FailedLogOnCount.Item1;
                    }

                    if (properties.LockedUntil != null)
                    {
                        accountData.LockedUntil = properties.LockedUntil.Item1;
                    }

                    if(properties.PasswordLastSetDate != null)
                    {
                        accountData.PasswordLastSetDate = properties.PasswordLastSetDate.Item1;
                    }
                    
                    break;

                }
            }


            SaveAllAccountDataToFile(accountDataList);
        }

        #endregion Public Methods

    }
}